-- -- --
--
-- Batch Craft Chat Filter addon
--  Corwin MacGregor
--
-- version 1.2f
-- 08/28/2013
--
-- -- --

--[[
--]]

---------------------
-- Local Variables --
---------------------
local addonName = ...
local addon  = LibStub("AceAddon-3.0"):NewAddon( addonName, "AceHook-3.0", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName)
local db
_G["BCCF"] = addon


local LOOT_ITEM_CREATED_SELF = LOOT_ITEM_CREATED_SELF
local LOOT_ITEM_CREATED_SELF_MULTIPLE = LOOT_ITEM_CREATED_SELF_MULTIPLE
local SKILL_RANK_UP = SKILL_RANK_UP

local alreadyBatching = false
local recipeID, recipeLink, recipeName, recipeQuality, recipeVerb
local tradeskillName
local recast = 0
local numMade = 0
local skillUps = {}
-- local expectingLoot = 0

local errorMsgs = {}
local lastEventArgs = {}
local filteredChatFrame

local timerID

local MIN_LINES = 2
local MAX_LINES = 128
local MIN_LOD = 1
local MAX_LOD = 4


---------------
-- DB default

local default_db = {
	global = {
		threshold = 2,
		color = {r=0.8, g=1.0, b=1.0},
		debug = {
			enabled = false,
			LoD = 0,	-- Level-of-Detail. 0 means all levels
		},
	},
}


---------------------
-- Output functions

addon.alias = "BCCF"

local printTypeInfo = {}
printTypeInfo["PRINT"]	= {r=1.0,g=1.0,b=1.0}
printTypeInfo["DEBUG"]	= {r=0.6,g=0.6,b=0.6}
printTypeInfo["ALERT"]	= {r=1.0,g=0.9,b=0.0}
printTypeInfo["ERROR"]	= {r=1.0,g=0.0,b=0.0}


local function RGBtoHex ( r, g, b, a )
	r = tonumber(format("%.4f", r)); g = tonumber(format("%.4f", g)); b = tonumber(format("%.4f", b));
	if r >= 0 and r <= 1 and g >= 0 and g <= 1 and b >= 0 and b <= 1 then
		-- rgb in percent form
		r = r * 255; g = g * 255; b = b * 255;
	elseif r >= 0 and r <= 255 and g >= 0 and g <= 255 and b >= 0 and b <= 255 then
		-- rbg in decimal form, no convertion necessary
	else
		error("E_RANGE: Malformed RGB values: "..strjoin(" ", "", tostringall(r, g, b, a)))
	end
	
	if not a then a = 255 end
	a = tonumber(format("%.4f", a))
	if a >= 0.0 and a <= 1.0 then a = a * 255
	elseif a >= 0 and a <= 255 then -- alpha already in decimal
	else error("E_RANGE: Malformed alpha value: "..strjoin(" ", "", tostringall(r, g, b, a)))
	end

	return strlower( format( "%02x%02x%02x%02x", a, r, g, b ) );	
end


local function color( text, ... )

	local color = select(1,...)
	if type(color) == "string" then
		-- we've received a hexcolor string
	elseif type(color) == "number" then
		-- we've prolly received numerics, make sure we have enough, parse to RGBtoHex()
		local r, g, b = ...
		if r and type(r) == "number" and g and type(g) == "number" and b and type(b) == "number" then
			color = RGBtoHex( r, g, b )
		else
			error("E_INVARG: Invalid or missing RGB values in call to color(): "..strjoin(" ", "", tostringall(...)))
		end
	elseif type(color) == "table" then
		-- we've received some sort of color table
		if #color == 3 then
			color = RGBtoHex( unpack(color) )
		elseif color.r and type(color.r) == "number"
		and color.g and type(color.g) == "number"
		and color.b and type(color.b) == "number" then
			color = RGBtoHex( color.r, color.g, color.b )
		else
			error("E_INVARG: Invalid or missing RGB values in table passed to color(): "..strjoin(" ", "", tostringall(...)))
		end
	end

	return format("\124c%s%s\124r", color, text)
end


local function tag_string ( text, printType )
	if not printType then error("E_ARGS: Missing printType on call to tag_string()") end

	local info = printTypeInfo[printType]
	if not info then error("E_INVARG: Invalid printType on call to tag_string()") end

	return format( "%s%s%s %s", color("[", db.color), color(addon.alias, info), color("]", db.color), text )
end


local function do_print ( text, printType )

	local info = db.color
	DEFAULT_CHAT_FRAME:AddMessage( tag_string( text, printType ), info.r, info.g, info.b )
end


local function printf ( ... ) do_print( format( ... ), "PRINT" ) end
local function alertf ( ... ) do_print( format( ... ), "ALERT" ) end
local function errorf ( ... ) do_print( format( ... ), "ERROR" ) end


--------------------
-- Debug functions

function addon.debuggable ( level )
	return db.debug.enabled and (db.debug.LoD >= level or db.debug.LoD == 0)
end


local function do_debugf ( level, ... ) do_print( format( "[Debug:%d] %s", level, format(...) ), "DEBUG" ) end


local function debugf ( level, ... )
	varargs = {...}
	if type(level) == "string" then	-- default debug level == 1
		varargs = {level, unpack( varargs )}
		level = 1
	end
	if addon.debuggable(level) then
		 do_debugf(level, unpack(varargs))
	end
end


-----------------------------
-- Options handlers & table

local function getOption (info)
	_t = db
	for _,v in ipairs(info) do
		_t = _t[v]
	end
	return _t
end


local function setOption (spoof, info, ...)
	-- setOption( [spoof,] info, ... )
	-- spoof must be a string, referencing what message to spoof
	-- info should always be the table passed by AceConfig to handler functions
	-- if spoofing, info must be modified to the correct keyvalue prior to setOption call

	local _v, _pv
	if type(spoof) ~= "string" and type(spoof) == "table" then
		_v = info; _pv = {info, ...}; info = spoof; spoof = nil;
	else
		_v = ...; _pv = {...};
	end

	_t = db
	_s = "OPTION_SETREPLY"
	for i = 1, (#info-1) do
		_t = _t[info[i]]
		_s = _s.."_"..info[i]
	end

	if info.type == "color" then
		for i,k in ipairs({"r", "g", "b", "a"}) do
			_t[info[#info]][k] = _pv[i]
		end
	else
		_t[info[#info]] = _v
	end
	_s = _s.."_"..(spoof or info[#info])

	if info.uiType == "cmd" then
		if info.type == "toggle" then
			if _v == true then
				_s = _s.."_T"
			elseif _v == false then
				_s = _s.."_F"
			elseif info.option.tristate and _v == nil then
				_s = _s.."_NIL"
			end
		end

		alertf( L[strupper(_s)], unpack(_pv) )
	end
end


local function getBoolOptionString (info, propName, value)
	if not value then
		-- try to get the value from the option's associated DB property
		value = getOption( info )
	end
	if type(value) == "nil" then error("E_INVARG: Missing boolean or database entry! "..strjoin(" ", info[#info], propName, value)) end

	_s = "OPTION_"..propName
	for i = 1, #info do _s = _s.."_"..info[i] end
	if value == true then
		_s = _s.."_T"
	elseif value == false then
		_s = _s.."_F"
	elseif info.option.tristate and value == nil then
		_s = _s.."_NIL"
	end

	return L[strupper(_s)]
end

local function getBoolOptionName (info, value) return getBoolOptionString (info, "NAME", value) end
local function getBoolOptionDesc (info, value) return getBoolOptionString (info, "DESC", value) end


local configOptions = {
	name = addonName,
	type = "group",
	get = getOption,
	set = setOption,
	args = {
		config = {
			name = "[PH] configuration",
			desc = L["OPTION_DESC_CONFIG"],
			type = "execute",
			guiHidden = true,
			func = function () InterfaceOptionsFrame_OpenToCategory(addon.optframe) end,
		},
		enable = {
			name = function (info) return getBoolOptionName(info,addon:IsEnabled()) end,
			desc = function (info) return getBoolOptionDesc(info,addon:IsEnabled()) end,
			type = "toggle",
			order = 2,
			get = function() return addon:IsEnabled() end,
			set = function() if addon:IsEnabled() then addon:Disable()
					else addon:Enable() end end,
		},
		version = {
			name = L["OPTION_NAME_VERSION"],
			type = "execute",
			guiHidden = true,
			get = false, set = false,
			func = function() alertf( L["OPTION_EXEC_VERSION"], GetAddOnMetadata( addonName, "Version" ) ) end,
		},
		threshold = {
			name = L["OPTION_NAME_THRESHOLD"],
			desc = L["OPTION_DESC_THRESHOLD"],
			type = "range",
			order = 3,
			min = MIN_LINES, max = MAX_LINES, step = 1, softMax = 8,
		},
		color = {
			name = L["OPTION_NAME_COLOR"],
			desc = L["OPTION_DESC_COLOR"],
			type = "color",
			order = 4,
		},
		debug = {
			name = L["OPTION_NAME_DEBUG"],
			type = "group",
			order = 5,
			guiInline = true,
			args = {
				enabled = {
					name = getBoolOptionName,
					desc = getBoolOptionDesc,
					type = "toggle",
					order = 5,
				},
				LoD = {
					name = L["OPTION_NAME_DEBUG_LOD"],
					desc = L["OPTION_DESC_DEBUG_LOD"],
					type = "range",
					order = 6,
					disabled = function() return not db.debug.enabled end,
					min = MIN_LOD, max = MAX_LOD, step = 1,
				},
				verbose = {
					name = L["OPTION_NAME_DEBUG_VERBOSE"],
					desc = L["OPTION_DESC_DEBUG_VERBOSE"],
					type = "execute",
					order = 7,
					disabled = function() return not db.debug.enabled end,
					get = false, set = false,
					func = function (info) info[#info]="LoD"; setOption("verbose",info,0) end,
				},
			},
		},
	},
}


-------------------
-- Addon Handlers

function addon:OnInitialize()

	-- setup database
	self.db = LibStub("AceDB-3.0"):New("BatchCraftChatFilterDB", default_db)
	db = self.db.global

	-- setup options
	LibStub("AceConfig-3.0"):RegisterOptionsTable( addonName, configOptions, {addonName, "bccf"} )
	self.optframe = LibStub("AceConfigDialog-3.0"):AddToBlizOptions( addonName )

end


function addon:OnEnable()
	self:Hook("DoTradeSkill", true)
end


function addon:OnDisable()
end


------------------
-- Miscellaneous

local function GetTradeSkillSpecInfo( index )
	local i = index
	repeat
		info = {GetTradeSkillInfo(i)}
		i = i - 1
	until (info[2] == "subheader" or info[2] == "header" or i < 1)

	if not info[8] then return end

	return info[1], info[9], info[10]
end


local function InitializeProfessionsTable()

	local profName, profCurrRank, profMaxRank = GetTradeSkillLine()
	tradeskillName = profName
	if profCurrRank < profMaxRank then
		skillUps[profName] = {["currRank"]=profCurrRank, ["maxRank"]=profMaxRank, ["dirty"]=nil}
	end

	local specName, specCurrRank, specMaxRank = GetTradeSkillSpecInfo( recipeID )
	if specName and specCurrRank < specMaxRank then
		skillUps[specName] = {["currRank"]=specCurrRank, ["maxRank"]=specMaxRank, ["dirty"]=nil}
	end

	debugf( 2, "profession=%q  specialty=%s", profName, (specName and format("%q", specName)) or "none")
	if addon.debuggable(2) then
		for _prof, _info in pairs(skillUps) do
			do_debugf( 2, "skillUps[%q]= %d / %d", _prof, _info.currRank, _info.maxRank )
		end
	end
end


local function Initialize()
	local self = addon
	debugf( "In Initialize()" )

	InitializeProfessionsTable()

	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("UNIT_SPELLCAST_FAILED")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:RegisterEvent("UI_ERROR_MESSAGE")
--	self:RegisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
--	self:RegisterEvent("TRADE_SKILL_CLOSE")
--	self:RegisterEvent("TRADE_SKILL_SHOW")

	ChatFrame_AddMessageEventFilter( "CHAT_MSG_LOOT", self.batchCraftLootFilter )
	if recipeQuality ~= "trivial" then
		ChatFrame_AddMessageEventFilter( "CHAT_MSG_SKILL", self.batchCraftSkillFilter )
	end
--	expectingLoot = 0
	numMade = 0

	errorMsgs = {}

	if recast > 1 then printf( L["MSG_BATCH_START"] ) end
	debugf( 2, "Exiting Initialize()" )
end


local function CleanUpSkill( forced )
	local self = addon
	debugf( "In CleanUpSkill() via %s...", forced and "timer" or "filter" )

	if timerID then debugf( 2, "Killing timer: %s", self:CancelTimer(timerID)) end

	ChatFrame_RemoveMessageEventFilter( "CHAT_MSG_SKILL", self.batchCraftSkillFilter )
	debugf( "skill-up filter removed." )

	if addon.debuggable(2) then
		local skillUpsLen = 0; for _ in pairs(skillUps) do skillUpsLen = skillUpsLen + 1 end
		do_debugf( 2, "skillUps: %d", skillUpsLen )
	end
	for skill, info in pairs(skillUps) do
		if info.dirty then
			debugf( 2, "skillUp: %q  rank( %d / %d )", skill, info.currRank, info.maxRank )
			lastEventArgs["SKILL"][1] = tag_string(format( SKILL_RANK_UP, skill, info.currRank ), "PRINT")
			ChatFrame_MessageEventHandler( filteredChatFrame, "CHAT_MSG_SKILL", unpack( lastEventArgs["SKILL"] ) )
			skillUps[skill] = nil
		end
	end

	local filters = ChatFrame_GetMessageEventFilters("LOOT")
	if filters then
		for index, filter in next, filters do
			if ( filter == addon.batchCraftLootFilter ) then return end
		end
	end
--	filteredChatFrame = nil
	timerID = nil
	alreadyBatching = false
	debugf( 2, "Exiting CleanUpSkill()" )
end
addon.CleanUpSkill = CleanUpSkill


local function CleanUp()
	local self = addon
	debugf( "In CleanUp()..." )

	ChatFrame_RemoveMessageEventFilter( "CHAT_MSG_LOOT", self.batchCraftLootFilter )
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:UnregisterEvent("UNIT_SPELLCAST_FAILED")
	self:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:UnregisterEvent("UI_ERROR_MESSAGE")
--	self:UnregisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
--	self:UnregisterEvent("TRADE_SKILL_CLOSE")
--	self:UnregisterEvent("TRADE_SKILL_SHOW")
	debugf( "loot filter removed, events unregistered." )

	if recast > 1 and tContains({"easy", "medium"}, recipeQuality) then
		debugf( 2, "Recipe quality is %q, preparing for lack of random proc on last cast...", recipeQuality )
		for skill, info in pairs(skillUps) do
			if info.dirty then
				debugf( 2, "Found dirty skill-up: %q", skill )
				timerID = self:ScheduleTimer( CleanUpSkill, 1, true )
				break;
			end
		end
	end

	debugf( 2, "numMade: %d", numMade )
	if numMade > 0 then
		if numMade == 1 then
			lastEventArgs["LOOT"][1] = tag_string(format( LOOT_ITEM_CREATED_SELF, recipeLink ), "PRINT")
		else
			lastEventArgs["LOOT"][1] = tag_string(format( LOOT_ITEM_CREATED_SELF_MULTIPLE, recipeLink, numMade ), "PRINT")
		end
		ChatFrame_MessageEventHandler( filteredChatFrame, "CHAT_MSG_LOOT", unpack( lastEventArgs["LOOT"] ) )
	end

	local filters = ChatFrame_GetMessageEventFilters("SKILL")
	if filters then
		for index, filter in next, filters do
			if ( filter == addon.batchCraftSkillFilter ) then return end
		end
	end
--	filteredChatFrame = nil
	alreadyBatching = false
	debugf( 2, "Exiting CleanUp()" )
end
addon.CleanUp = CleanUp


local function PatternizeGlobalString( text )
	local replacements = {["%."]="%%%.", ["%%s"]="(.*)", ["%%d"]="(%%d*)"}
	for k,v in pairs(replacements) do
		text = gsub( text, k, v )
	end
	return "^"..text.."$"
end
local LICS_SINGLE_PATTERN  = PatternizeGlobalString( LOOT_ITEM_CREATED_SELF )
local LICS_MULTIPLE_PATTERN = PatternizeGlobalString( LOOT_ITEM_CREATED_SELF_MULTIPLE )
local SRU_PATTERN = PatternizeGlobalString( SKILL_RANK_UP )
if SKILL_RANK_UP ~= ERR_SKILL_UP_SI then errorf(L["MSG_ERR_SKILL_STRING_ASYNO"], GetLocale()) end


local function GetAcceptableNumCraftingSpamLines ()
	-- do some sort of calculations to figure this out
	-- otherwise, possibly by setting
	-- until then, we're gonna just make it a flat "4"
	-- v1.2: get value from db

	return db.threshold
end


function addon:Abort_Handler ( event, unit )
	if unit ~= "player" then return end

	debugf( "Aborting via %s from %q.  Recast set to zero.  Going to clean-ups.", event, unit )

	recast = 0
	CleanUp();  -- schedule loot filter removal
	CleanUpSkill();  -- schedule skill-up filter removal
end


-------------------
-- Event handlers

function addon:UNIT_SPELLCAST_SUCCEEDED ( event, unit, spell )
	if unit ~= "player" then return end

	debugf( "Succeeded(): "..spell )

--	expectingLoot = expectingLoot + 1
	if spell ~= recipeName then
		debugf( "Whoops. This is not the spell we were expecting.  bail." )
		return
	end
	recast = recast - 1
	recipeQuality = select(2, GetTradeSkillInfo(recipeID))
	debugf( "Reduced recast by 1. updating recipeQuality. Exiting Succeeded()." )
end


function addon:UNIT_SPELLCAST_FAILED ( event, unit )
	if unit ~= "player" then return end

	debugf( "Failed(). Going to Abort_Handler()." )

	self:Abort_Handler( event, unit )
	errorf( L["MSG_BATCH_FAILED"], errorMsgs[#errorMsgs] )
--	CleanUp();  -- schedule loot filter removal
end


--[[ -- apparently this event doesn't fire when interrupting casting with movement
--   -- since UNIT_SPELLCAST_INTERRUPTED does, we'll go with that for both conditions
function addon:UNIT_SPELLCAST_FAILED_QUIET ( event, unit )
	if unit ~= "player" then return end

	debugf( "Failed_Quiet().  Going to Abort_Handler().  Listening for UNIT_SPELLCAST_INTERRUPTED" )

	self:Abort_Handler( event, unit )
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
end
--]]


function addon:UNIT_SPELLCAST_INTERRUPTED ( event, unit )
	if unit ~= "player" then return end

	debugf( "Interrupted().  Going to Abort_Handler()." )

	alertf( L["MSG_BATCH_INTERRUPTED"] )
	self:Abort_Handler( event, unit )	
--	CleanUp();	-- schedule loot filter removal
end


function addon:UI_ERROR_MESSAGE ( event, msg )
	debugf( "Error Message received.  Something happened..." )
	debugf( 2, "%q error caught in :UI_ERROR_MESSAGE()", msg )

	tinsert( errorMsgs, msg )
end


-- these don't seem to work.  how annoying.
-- function addon:TRADE_SKILL_CLOSE ( ... )  self:Abort_Handler( ... )  end
-- function addon:TRADE_SKILL_SHOW ( ... )  self:Abort_Handler( ... )  end


-----------
-- Filter

function addon.batchCraftLootFilter( chatFrame, event, msg, ... )

	debugf( "In loot filter...")
	debugf( 2, "event=%s  msg=%q", event, msg )

--[[  -- the "Expecting loot msg" system seemed to crash badly on split-stack loot situations
	-- disabled until bug can be logically resolved
	if expectingLoot >= 1 then           -- we had a SUCCEEDED, so we're expecting this LOOT msg
		expectingLoot = expectingLoot - 1
	else
		-- some sort of sync error
		errorf("E_SYNC: sync error in filter")
		return false, msg, ...;
	end
--]]

	local retval = {false, msg, ...}

	if filteredChatFrame and not( filteredChatFrame==chatFrame ) then
		return unpack(retval)
	end
	filteredChatFrame = chatFrame
	lastEventArgs["LOOT"] = {msg, ...}

	local link, amount

	debugf( 2, "Attempting match against %q", msg )
	for _,p in ipairs({LICS_MULTIPLE_PATTERN, LICS_SINGLE_PATTERN}) do
		debugf( 2, "Attempting match with %q", p )
		link, amount = strmatch( msg, p )
		debugf( 2, (link and "match found: link=%q  amount=%d") or "match not found", link or "nil", ((not link and -1) or (amount or 1)) )
		if link then	-- we've had a match
			debugf( 2, "We've had a match: %s x%d", link, (amount or 1) )
			numMade = numMade + (amount or 1)
			debugf( 2, "Accumulation: %sx%d", link, numMade )
			retval = {true}
			break
		end
	end

	debugf( 2, "LOOT msg schedule removal?  recasts: %d", recast )
	if (not TradeSkillFrame:IsShown()) or (CURRENT_TRADESKILL ~= tradeskillName) or (recast <= 0) then
		if addon.debuggable(2) then
			if recast <= 0 then
				do_debugf( 2, "Scheduling CleanUp() because we've run out of recasts." )
			elseif not TradeSkillFrame:IsShown() then
				do_debugf( 2, "Scheduling CleanUp() because TradeSkillFrame not :IsShown()" )
			elseif (CURRENT_TRADESKILL ~= tradeskillName) then
				do_debugf( 2, "Scheduling CleanUp() because Tradeskill has changed." )
			end
		end
		CleanUp()		-- schedule filter removal
	end

	debugf( "Exiting loot filter." )
	return unpack( retval )

end


function addon.batchCraftSkillFilter( chatFrame, event, msg, ... )
	debugf( "In skill filter..." )
	local recast = recast  -- cache this locally so last LOOT doesn't kill it by the time we get to end of this code

	local retval = {false, msg, ...}

	if filteredChatFrame and not( filteredChatFrame==chatFrame ) then
		return unpack(retval)
	end
	filteredChatFrame = chatFrame
	lastEventArgs["SKILL"] = {msg, ...}

	debugf( 2, "Attempting match against %q", msg )
	local skill, rank = strmatch( msg, SRU_PATTERN )
	debugf( 2, "match found: skill=%q  rank=%d", skill or "nil", rank or -1 )
	if skill and skillUps[skill] then	-- we've had a match, and we're expecting it
		skillUps[skill].currRank = rank
		skillUps[skill].dirty = true
		retval = {true}
	end
	if addon.debuggable(2) then
		for _prof, _info in pairs(skillUps) do
			do_debugf( 2, "skillUps[%q]= %d / %d", _prof, _info.currRank, _info.maxRank )
		end
	end

	debugf( 2, "SKILL msg schedule removal?  recasts: %d", recast )
	if (recast <= 0) then
		CleanUpSkill()	-- schedule filter removal
	end

	debugf("Exiting skill filter.")
	return unpack( retval )
end


---------
-- Hook

function addon:DoTradeSkill( index, numRepeat )

	debugf( "in DoTradeSkill() hook...")
	if alreadyBatching then debugf( "Exiting superfluous hook." ); return end
	alreadyBatching = true
	debugf( 2, "index=%d  numRepeat=%d", index, numRepeat )

	if numRepeat >= GetAcceptableNumCraftingSpamLines() then
		debugf( 2, "Suppressing spam..." )
		recast = numRepeat
		recipeID = index
		recipeLink = GetTradeSkillItemLink( index )
		recipeName, recipeQuality, _, _, recipeVerb = GetTradeSkillInfo( index )
		if recipeVerb then
			debugf( "Exiting hook via altVerb clause." ); return
		end
		debugf( 2, "link=%s  recipe=%q  quality=%q", recipeLink, recipeName, recipeQuality )
		Initialize()
	end
	debugf( "Exiting hook." )
end



-- // --